/*
Node that segments an input frame and publishes the binary mask image in ROS. Using Ajay's active segmentation (2D)
If necessary, segments objects from PCL data and saves PCL data into .pcd file (basically a filtering operation)
Author: Ching L. Teo
Date: 06/01/2011
*/

// ROS core
#include <ros/ros.h>
#include <boost/thread/mutex.hpp>
#include <image_transport/image_transport.h>
#include <cv_bridge/cv_bridge.h>
#include <sensor_msgs/image_encodings.h>
//openCV
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/core/core.hpp>
// PCL includes
#include <pcl_ros/point_cloud.h>
#include <pcl/io/io.h>
#include <pcl/io/pcd_io.h>
#include <pcl/point_types.h>
#include <pcl/range_image/range_image_planar.h>
#include <pcl/filters/passthrough.h>
#include <pcl/common/common_headers.h>
#include <pcl_visualization/pcl_visualizer.h>
// others
#include <segmentationLayer.h>
#include <stdio.h>
#include <algorithm>

using namespace std;
using namespace pcl;
using namespace pcl_visualization;
using namespace cv;
namespace enc = sensor_msgs::image_encodings;
static const char WINDOW[] = "Input";
static int NUM_FRAMES_PROC = 0;
static const char SAVE_DIR[] = "/media/Data_Large/filter-training/";
static const char OBJ_NAME[] = "test";
static const int FIX_X = 140;
static const int FIX_Y = 206;
static const int SHIFT_X = -5;
static const int SHIFT_Y = 11;

class SimpleSegment2D
{

public:
	SimpleSegment2D(int argc, char** argv) : it(node_handle)
	{
		// check command line inputs
		ProcessCommandLine(argc, argv);

		//check input topic
		if (node_handle.resolveName("input")=="/input")
		{	
    			std::cerr << "Did you forget input:=<your topic>?\n";
			exit(0);
		}

		// set-up attributes + subscriber
		segL = new segLayer();
		old_proc_width = proc_width = 320*2; // default processing size
		old_proc_width = proc_height = 240*2;
		bSkip = false;
		bDemoMODE = true;
		bFilterTrainMODE = false;
		bPCLTrainMODE = false;
		bShowPCLViewer = false;
		iNumTrainImg = 5; 
		if(bFilterTrainMODE || bPCLTrainMODE){
			char cmdstr[100];
			sprintf(cmdstr, "mkdir %s", SAVE_DIR);		
			int retval = system(cmdstr);
			if(bShowPCLViewer)
				viewer = new PCLVisualizer("Live viewer - point cloud");
		}		
		if(bDemoMODE)
        	bRedoEdgeDet = true; // hack for static scenes, do not redo the edge computations
		else
			bRedoEdgeDet = false;
		fTol = 0.3f;
		subscriber_img = it.subscribe ("input", 10, &SimpleSegment2D::image_msg_cb, this);
		//subscriber_depth = it.subscribe ("/umd_image_proc/image_umd_processKinectRange2", 1, &SimpleSegment2D::depthimg_cb, this); // receive depth images from process Kinect node
		
		subscriber_depth = it.subscribe ("/camera/depth/image", 1, &SimpleSegment2D::depthimg_cb, this); // receive depth images
		sub_pcl_rgb = node_handle.subscribe("/camera/rgb/points", 1, &SimpleSegment2D::cloud_msg_cb, this);		

		//set-up publishers
		publisher_img = it.advertise("/umd_image_proc/image_umd_SimpleSegment2D", 1);
		publisher_pb = it.advertise("/umd_image_proc/image_umd_SimpleSegment2D_pb", 1);
		pub_pcl_rgb = node_handle.advertise <PointCloud<PointXYZRGB> >("/umd_image_proc/pcl_objseg", 1);


	}
	
	~SimpleSegment2D()
	{
		if(segL!=NULL){	
			delete segL;		
		}
	}

	//callback for subscriber - extract image messages (fast)
	void image_msg_cb (const sensor_msgs::ImageConstPtr& msg)
	{
		m.lock ();
		image_msg = msg;
		m.unlock ();

		ProcessMessage();

	}

	//callback for subscriber - extract depth image (slow)
	void depthimg_cb (const sensor_msgs::ImageConstPtr& msg)
	{
		m.lock ();
		depthimg_msg = msg;
		m.unlock ();

		ProcessMessage();

	}

	// call backs for subscriber
	void cloud_msg_cb (const sensor_msgs::PointCloud2ConstPtr& msg)
	{
	  m.lock ();
	  cloud_msg = msg;
	  //std::cout << "Received cloud msg: " << "header=" << msg->header << "width="<<msg->width <<", height="<<msg->height<<".\n";
	  m.unlock ();
	 
	  ProcessMessage();
	}

private:
	// ros attributes
	ros::NodeHandle node_handle;
	image_transport::ImageTransport it; // should always use ImageTransport to subscribe and publish images
	image_transport::Subscriber subscriber_depth;
  	image_transport::Subscriber subscriber_img;
  	image_transport::Publisher publisher_img;
	image_transport::Publisher publisher_pb;
	sensor_msgs::ImageConstPtr image_msg, old_image_msg;
	sensor_msgs::ImageConstPtr depthimg_msg, old_depthimg_msg;
	ros::Publisher pub_pcl_rgb; // publishes the segmented PCL data (for visualization)
	ros::Subscriber sub_pcl_rgb;
	boost::mutex m;
	sensor_msgs::PointCloud2ConstPtr cloud_msg, old_cloud_msg;
	PCLVisualizer* viewer;		

	// other
	segLayer* segL;
	int proc_width, old_proc_width;
	int proc_height, old_proc_height;
	bool bSkip;
    bool bRedoEdgeDet;
	bool bDemoMODE;
	bool bFilterTrainMODE; // set to true to collect filter training data
	int iNumTrainImg; // number of training images to collect;
	bool bPCLTrainMODE; // set to true to save PCL training date
	bool bShowPCLViewer;
	float fTol; // tolerance between image and depth frames time stamps 


	
	// Main internal function to process the message
	void ProcessMessage()
	{

		if(bPCLTrainMODE && bShowPCLViewer){
			usleep (100000);
			viewer->spinOnce (10);
		}
		// if no new messages appear, just stop processing
		if (!image_msg || image_msg == old_image_msg || !depthimg_msg || depthimg_msg == old_depthimg_msg || !cloud_msg || cloud_msg == old_cloud_msg)
			return;
		
		// if headers do not match, return
		ros::Time imgTime = image_msg->header.stamp;
		ros::Time depthTime = depthimg_msg->header.stamp;
		ros::Duration diffTime = (imgTime - depthTime);
		if( abs(diffTime.nsec*pow(10,-9)) > fTol ){
			cout << " img: "<<image_msg->header.stamp;
			cout << " depth: "<< depthimg_msg->header.stamp;
			cout << " diff: "<< abs(diffTime.nsec*pow(10,-9));
			ROS_WARN("header mismatch\n");
			return;
		}
		old_image_msg = image_msg;
		old_depthimg_msg = depthimg_msg;
		old_cloud_msg = cloud_msg;

		// convert message from ROS to openCV
		cv_bridge::CvImagePtr cv_ptr, cv_depthptr;
		try
		{
			cv_ptr = cv_bridge::toCvCopy(image_msg, enc::BGR8);
			cv_depthptr = cv_bridge::toCvCopy(depthimg_msg, enc::TYPE_32FC1);
		}
		catch (cv_bridge::Exception& e)
		{
			ROS_ERROR("cv_bridge exception: %s", e.what());
			return;
		}
	
		
		// convert to IplImage
		IplImage ipl_img = cv_ptr->image;
		IplImage ipl_depth = cv_depthptr->image;

		ROS_INFO("(img_h, img_w)=%d,%d.\n", ipl_img.height, ipl_img.width);     
		ROS_INFO("(depth_h,depth_w)=%d,%d.\n", ipl_depth.height, ipl_depth.width); 
				
		// check if the previous frame had the same width and height, if not, wait for 1 frame
		if(proc_width != old_proc_width || proc_height != old_proc_height)
		{
			ROS_WARN("Disparity map size change");
			bSkip = true;
		}
		else{
			bSkip = false;	
			if(segL==NULL)
				segL = new segLayer();
		
		}	
		//bSkip = false; // hack to show why not resetting causes failure in cvMul	
		old_proc_width = proc_width;
		old_proc_height = proc_height;

		if(bFilterTrainMODE && NUM_FRAMES_PROC > iNumTrainImg){
			ROS_ERROR("Finished Filter Training\n");
			exit(0);
		}

		if(!bSkip){
			// reduce to desired size
			IplImage* ipl_img_small = cvCreateImage(cvSize(proc_width, proc_height), IPL_DEPTH_8U, 3);
			cvResize(&ipl_img, ipl_img_small);

			// make sure depth image is the same size		
			IplImage* ipl_depth_small = cvCreateImage(cvSize(proc_width, proc_height), IPL_DEPTH_32F, 1);
			cvResize(&ipl_depth, ipl_depth_small);	

			//DEBUG
			double minval, maxval;
			cvMinMaxLoc(&ipl_depth, &minval, &maxval);
			ROS_INFO("min: %g, max: %g", minval, maxval);
			//cvSaveImage("depthmap.png", &ipl_depth);
					
///* enable this block comment to show that it is segLayer() mem leak			
			// set in the image and disparity map
			segL->setImage(ipl_img_small);		

			if(segL->setDisparity(ipl_depth_small)==1){
				ROS_ERROR("Disparity map not loaded correctly");
				return;
			}

			if(bDemoMODE && bRedoEdgeDet){
				// Edge detection
			  	segL->edgeCGTG();
				segL->generatePbBoundary();
				bRedoEdgeDet = false; // do once only
			}
			// always do Edge detection
			if(!bDemoMODE){
				segL->edgeCGTG();
				segL->generatePbBoundary();
			}
		
			// just give a fixed point to segment
			segL->assignFixPt(FIX_X, FIX_Y);

			if(bDemoMODE)            
				segL->selectFixPt_interactive();

			//segment
			segL->allocateMemForContours();
			segL->segmentCurrFix();

			// save object masks
			char prefix[100];
			ros::Time tstamp=ros::Time::now();
			double dSegTimeStamp = tstamp.toSec(); 
			if(bFilterTrainMODE && NUM_FRAMES_PROC <= iNumTrainImg){
				sprintf(prefix, "%s%s_%f", SAVE_DIR, OBJ_NAME, dSegTimeStamp);
				//char* prefix = "/home/erratic/filter-training/tomato1";
				//segL->saveRegions(prefix);
				segL->saveRegionsnRGB(prefix);
			}
			
			// the target point cloud that we will fill segmented PCL data with
			// get message from ROS and convert to point cloud
			PointCloud<PointXYZRGB> point_cloud;
			fromROSMsg(*cloud_msg, point_cloud);
			PointCloud<PointXYZRGB>::Ptr msg_orig_rgb (new PointCloud<PointXYZRGB>(point_cloud)); // ptr to original RGB point cloud
			PointCloud<PointXYZRGB>::Ptr msg_rgb (new PointCloud<PointXYZRGB>); // ptr to segmented point cloud
			
			IplImage* img_mask = segL->getCurrMask();
				//cvNamedWindow("mask",1);
				//cvSaveImage("mask.png",img_mask);
			if(bPCLTrainMODE && NUM_FRAMES_PROC <= iNumTrainImg){	
							
				createSegPCL(msg_orig_rgb, msg_rgb, img_mask, ipl_img_small);	

				//publish segmented point cloud
				pub_pcl_rgb.publish((*msg_rgb));

				if(bShowPCLViewer){
					//Visualize -- DEBUG
					viewer->removePointCloud ("range image cloud");
					// display RGB point cloud
					pcl_visualization::PointCloudColorHandlerRGBField<pcl::PointXYZRGB> color_handler_cloud(*msg_rgb); 
					viewer->addPointCloud (*msg_rgb, color_handler_cloud, "range image cloud");	
				}
				char pcdName[100];
				char pcdNameFull[100];	
				ros::Time tstamp=ros::Time::now();
				double dSegTimeStamp = tstamp.toSec(); 
				sprintf(pcdName, "%s%s_%f.pcd", SAVE_DIR, OBJ_NAME, dSegTimeStamp);
				sprintf(pcdNameFull, "%s%s_orig_%f.pcd", SAVE_DIR, OBJ_NAME, dSegTimeStamp);

				pcl::PCDWriter writer;
				writer.write(pcdName, *(msg_rgb), false);
				writer.write(pcdNameFull, *(msg_orig_rgb), false);

				//pcl::io::savePCDFile(pcdName, *(msg_rgb), false); // save in binary mode
				//pcl::io::savePCDFile(pcdNameFull, *(msg_orig_rgb), false); // save in binary mode

  				//std::cout << "There are " << msg_rgb->points.size () << " data points to test_pcd.pcd." << std::endl;
				//std::cout << "There are " << msg_rgb->height << " height"<< msg_rgb->width <<" width to test_pcd.pcd." << std::endl;

			}
	
//*/
			// get current Pb and seg res image results 
			IplImage* pbres = segL->getCurrPb();
			IplImage* res = segL->getCurrSegImg();
			if (res == NULL){
				ROS_WARN("no segmentation results");
				

				// check if publishing messages caused the leak
				cv_bridge::CvImage cvres_out;
				convertIpl2CvImage(ipl_img_small, &cvres_out, enc::BGR8);
				publisher_img.publish(cvres_out.toImageMsg());
				convertIpl2CvImage(ipl_depth_small, &cvres_out, enc::MONO8);
				publisher_pb.publish(cvres_out.toImageMsg());

				// still need to release
				//segL->deallocateMemForContours();
				cvReleaseImage(&ipl_img_small);
				cvReleaseImage(&ipl_depth_small);

				return;
			}
	
			// convert + publish results
			cv_bridge::CvImage cvres_out;
			convertIpl2CvImage(res, &cvres_out, enc::BGR8);
			publisher_img.publish(cvres_out.toImageMsg());
			convertIpl2CvImage(pbres, &cvres_out, enc::MONO8);
			publisher_pb.publish(cvres_out.toImageMsg());

			//display
			//segL->displayCurrSegs(10);
			
			//release memory
			segL->deallocateMemForContours();
			cvReleaseImage(&res);
			cvReleaseImage(&pbres);

			cvReleaseImage(&ipl_img_small);
			cvReleaseImage(&ipl_depth_small);
			cvReleaseImage(&img_mask);

		}
		else
		{
			// reset			
			if(segL!=NULL)	{
				delete segL;
				segL = NULL;
			}
		}

		NUM_FRAMES_PROC++;
		
	}

	void convertIpl2CvImage(IplImage* res, cv_bridge::CvImage* cvres_out, const std::string enc_type)
	{
		// convert to cv::Mat
		cv::Mat resMat(res);		

		// copy and publish
		cvres_out->header = image_msg->header;
		cvres_out->encoding = enc_type;
		cvres_out->image = resMat;
		
	}
	
	void createSegPCL(PointCloud<PointXYZRGB>::Ptr& msg_orig_rgb, PointCloud<PointXYZRGB>::Ptr& msg_rgb, IplImage* img_mask, IplImage* img_rgb)
	{


		IplImage* segRGB = cvCreateImage(cvSize(proc_width, proc_height), IPL_DEPTH_8U, 3);
		//IplImage* segDepth = cvCreateImage(cvSize(proc_width, proc_height), IPL_DEPTH_32F, 1);
		//cvSetZero(segDepth);
		cvSetZero(segRGB);
		cvCopy(img_rgb, segRGB, img_mask);
		//cvCopy(img_depth, segDepth, img_mask);

		//DEBUG
		cvSaveImage("seg_rgb.png",segRGB);
		

		//cvReleaseImage(&segRGB);
		//cvReleaseImage(&segDepth);


		//DEBUG
		//cvNamedWindow("mask_img", 1);
		//cvShowImage("mask_img", img_mask);
		//cvWaitKey(-1);

		// sanity check
		if (img_mask->width != msg_orig_rgb->width || img_mask->height != msg_orig_rgb->height){
			ROS_ERROR("Point Cloud size mismatch with mask\n");
			//TODO: resize point cloud or return?	
			exit(1);				
		
		}

		// copy the point cloud data
		//pcl::copyPointCloud(*(msg_orig_rgb), *(msg_rgb));
			
	
		//Populate PCL RGB+DEPTH Object

		msg_rgb->header.frame_id = "/openni_rgb_optical_frame";
		//msg_rgb->height = msg_orig_rgb->height;
		//msg_rgb->width = msg_orig_rgb->width;
		//msg_rgb->is_dense = true;
		//msg_rgb->points.resize(msg_orig_rgb->height*msg_orig_rgb->width);
	
		for (int y=0; y<msg_orig_rgb->height; y++)
		{
			for(int x=0;x<msg_orig_rgb->width;++x)
			{
				double val = cvGet2D(img_mask, y, x).val[0];
				CvScalar cv_rgbval = cvGet2D(segRGB, y, x); // get rgb value...
				//printf("%f\n",val);
				// copy data of point cloud 

				if (val > 0.0f){

					// get a sample data point from the point cloud
					PointXYZRGB tpoint;			
					tpoint = msg_orig_rgb->at(x+SHIFT_X,y+SHIFT_Y); // sample pcl point here
					//std::vector<uint8_t> rgb_t(4); // hold 4 elements 
					//extract_pcl_rgb(tpoint, rgb_t); 

					uint8_t r = (uint8_t)cv_rgbval.val[2]; 
                	uint8_t g = (uint8_t)cv_rgbval.val[1]; 
                	uint8_t b = (uint8_t)cv_rgbval.val[0];           
                	uint32_t rgb_val = ((uint32_t)r << 16 | (uint32_t)g << 8 | (uint32_t)b); // set to the appropriate value
                	tpoint.rgb = *reinterpret_cast<float*>(&rgb_val);

					// copy this data into the new point cloud -- OVERWRITE, not ADD
					msg_rgb->push_back(tpoint);
					//msg_rgb->points[(y+SHIFT_Y)*msg_orig_rgb->width+(x+SHIFT_X)] = tpoint;
					// x, y, z in meters. x right, y down, z forward from camera center, r, g, b:(1 - 255)
					//ROS_INFO("(x,y,z)=%g,%g,%g, (r,g,b)=%d,%d,%d", tpoint.x, tpoint.y, tpoint.z, rgb_t[0],rgb_t[1],rgb_t[2]); 
						
				}
/*
				else{
					PointXYZRGB tpoint;
					tpoint = msg_orig_rgb->at(x,y);
					//std::vector<uint8_t> rgb_t(4); // hold 4 elements 
					//extract_pcl_rgb(tpoint, rgb_t);
					uint8_t r = 0; 
                	uint8_t g = 0; 
                	uint8_t b = 0;           
                	int32_t rgb_val = (r << 16) | (g << 8) | b; // set to zero
                	tpoint.rgb = *(float *)(&rgb_val); 
					msg_rgb->points.push_back(tpoint);
				}
*/
/*
				if (val == 0.0f){
					PointXYZRGB tpoint;
					tpoint = msg_orig_rgb->at(x,y);
					//std::vector<uint8_t> rgb_t(4); // hold 4 elements 
					//extract_pcl_rgb(tpoint, rgb_t);
					uint8_t r = 0; 
                	uint8_t g = 0; 
                	uint8_t b = 0;           
                	int32_t rgb_val = (r << 16) | (g << 8) | b; // set to zero
                	tpoint.rgb = *(float *)(&rgb_val); 
					msg_rgb->points[y*msg_orig_rgb->width+x] = tpoint;
		    	}
*/
			}
		}
			
	}

	// function to extract rgb stored in pcl PointXYZRGB
	void extract_pcl_rgb(pcl::PointXYZRGB& point_t, std::vector<uint8_t>& rgb_v)
	{
		// extract color values
		uint32_t rgb_val_;
		memcpy(&rgb_val_, &(point_t.rgb), sizeof(float));

		uint8_t garbage_ = (uint8_t)((rgb_val_ >> 24) & 0x000000ff);
		uint8_t r_ = (uint8_t)((rgb_val_ >> 16) & 0x000000ff);
		uint8_t g_ = (uint8_t)((rgb_val_ >> 8) & 0x000000ff);
		uint8_t b_ = (uint8_t)((rgb_val_) & 0x000000ff);

		// save to rgb vector
		rgb_v[0]=r_; 
		rgb_v[1]=g_; 
		rgb_v[2]=b_; 
		rgb_v[3]= garbage_;

	}

	void ProcessCommandLine(int argc, char** argv)
	{
	}

	void printUsage (const char* progName)
	{
	}
};

int main(int argc, char** argv)
{

	ros::init (argc, argv, "umd_activeSeg2D");
	SimpleSegment2D simpseg(argc, argv);
	ros::spin();
	return 0;
  
}
