/********************************************************/
/* Create the polar edge map  			     	*/
/* Author:   		Ajay K Mishra      		*/
/* Modified On :  	Nov 02, 2010  		        */
/* Modified On :   	Feb 17, 2010 (Multi-threaded)	*/
/* email id: 		mishraka@gmail.com 		*/
/********************************************************/
// For Multi-threaded, collect the total number of edge pixels and divide it by number of available cores
// Run that many threads. 

#define _NUM_CORES 8
#include <cmath>
#include <pthread.h>
#include <vector>
#include <cstdlib>
#include <cstdio>

#define _PI acos(-1.0)
#define GAUSS2D(xdiff, ydiff, var) exp(-((xdiff*xdiff)/(2*var*var)+(ydiff*ydiff)/(4)))
static double maxIntensity=0; 

/*********************************/
/* Calculate circular histogram  */
/*********************************/
void calc2DHist(int origin_x, 
			 int origin_y, 
			 int edgept_x,
			 int edgept_y,
			 double grad, 
			 double ori,
			 const int width,
			 const int height, 
			 double* rThetaPlot_ptr){

	int nbins  = height;
	int nrbins = width;
	
	//each edge point contributes to all the bins
	double delx = -origin_x + edgept_x;
	double dely = -origin_y + edgept_y;
	double dist = sqrt(delx*delx+dely*dely);
	
	// center point and edgept should not be the same
	if(dist == 0 || ceil(dist) >= width-5) 
		return;
    	double theta=(atan2(-dely, delx)*(180/_PI));
    	theta = (theta<0) ? 360+theta :theta;
  
	double var  = 90.0/pow(dist, 0.7); 
    
    	int intv = var < 4 ? 4 :var;     
    	int min_bin  =  ceil(theta)- intv;
    	int max_bin  =  ceil(theta)+ intv;
    	int max_rbin =  ceil(dist) + intv;
    	int min_rbin =  ceil(dist) - intv;
    
	double theta_t;
    	double theta_a       = (_PI - ori);
    	double theta_n ;   
    
	min_rbin = (min_rbin <=1) ? 1 :min_rbin;
	max_rbin = (max_rbin >= width)? width-1 : max_rbin;
	double varAlongMajorAxis;
    	for (int rbin=min_rbin; rbin <= max_rbin; rbin++){
        	for(int bin=min_bin;bin<max_bin;bin++){
            	int binInd;
            	if(bin < 0){
                	binInd =bin+360;
            	}
            	else if(bin >= 360){
                	binInd =bin-360;
            	}
            	else{
                	binInd=bin;
            	}
            
            	if (binInd > 180){
                	theta_t   = (binInd-360)*_PI/180;
            	}
            	else{
                	theta_t   = binInd*_PI/180;
            	}            
            	theta_n       = theta_a - theta_t;
            	theta_n       = theta_n < 0 ? theta_n + _PI : theta_n;
		theta_n       = theta_n >= _PI ? theta_n - _PI : theta_n;
		double u_n    = cos(theta_n); 
            	double v_n    = sin(theta_n);
            
            	//find delX, and delY
            	double xChng			=	(rbin*cos(binInd*_PI/180))-delx;
            	double yChng			=	(rbin*sin(binInd*_PI/180))+dely;

            	double xChng_new        		=   xChng*cos(-ori)+yChng*sin(-ori);
            	double yChng_new        		=  -xChng*sin(-ori)+yChng*cos(-ori);

	    	double polar_angle_deg = u_n != 0 ? atan(v_n/u_n) : _PI/2.0 ;            
	    	varAlongMajorAxis   = 1.0*(1+fabs(polar_angle_deg)*180/_PI)/pow(dist, 0.7*fabs(sin(polar_angle_deg)));

            	*(rThetaPlot_ptr + width*binInd + rbin) += grad*GAUSS2D(xChng_new, yChng_new, varAlongMajorAxis);

		maxIntensity = (maxIntensity < *(rThetaPlot_ptr + width*binInd + rbin) ) ? *(rThetaPlot_ptr + width*binInd + rbin) : maxIntensity;
        }
    }	
}	

//**********************************************************************//
// -----------------MULTITHREADING--------------------------------------// 
//**********************************************************************//
struct threadArg_calc2DHist{
	int* x;
	int* y;
	int origin_x,origin_y;
	double* grad;
	double* ori;
	int width;
	int height;
	double* rThetaPlot_ptr;
	int startInd;
	int endInd;
};

void* threadFn_calc2DHist(void* arg){
	struct threadArg_calc2DHist* tmpArg = (struct threadArg_calc2DHist*)arg;

	for(int i= tmpArg->startInd; i<tmpArg->endInd; i++){
		//call calc2D function here!
		calc2DHist( tmpArg->origin_x, 
			    tmpArg->origin_y,
		    	    tmpArg->x[i],
		    	    tmpArg->y[i],
		            tmpArg->grad[i],
		            tmpArg->ori[i],
			    tmpArg->width,
			    tmpArg->height,
		            tmpArg->rThetaPlot_ptr);
	}
	//Kill the thread!
	fprintf(stdout," ."); fflush(stdout); 
	pthread_exit(NULL);	 	 

}


/******************************************************/
/* Generate polar plot using directional procession!! */
/******************************************************/
void genRThetaPlot( double* edgeGrad_ptr,
		    double* edgeOri_ptr,
                    double* rThetaPlot_ptr, 
		    const int origin_x,
		    const int origin_y,
		    const int width_polar,
		    const int height_polar,
		    const int width_cart,
		    const int height_cart){	    

	//create as many rThetaPlots for each thread
	double* tmpRThetaPlot_ptr[_NUM_CORES];
	for(int i=0; i<_NUM_CORES; i++)
		tmpRThetaPlot_ptr[i] = (double*) malloc(sizeof(double)*width_polar*height_polar);
 
	// initialize the rThetaPlots to zero
	for(int i=0; i < height_polar; i++)
		for(int j=0; j < width_polar; j++){			
			*(rThetaPlot_ptr + width_polar*i + j) = 0;
			for(int t=0; t < _NUM_CORES; t++)		
				*(tmpRThetaPlot_ptr[t] + width_polar*i + j) = 0; 
		}

	std::vector<int> edgePts_x;
	std::vector<int> edgePts_y;
	std::vector<double> edgePts_grad;
	std::vector<double> edgePts_ori;

	for (int x=0; x < width_cart; x++){
		for (int y=0; y < height_cart; y++){
			if (*(edgeGrad_ptr + width_cart*y + x) > 5.0/255){
				edgePts_x.push_back(x);
				edgePts_y.push_back(y);
				edgePts_grad.push_back(*(edgeGrad_ptr + width_cart*y + x));
				edgePts_ori.push_back(*(edgeOri_ptr + width_cart*y + x));				
			}
		}
	  }


	int threadWidth       = floor(edgePts_x.size()/_NUM_CORES);
	int nThreads          = _NUM_CORES;
	
	//start _NUM_CORES threads
	pthread_t* pth = (pthread_t *)malloc(nThreads*sizeof(pthread_t));
	struct threadArg_calc2DHist** tmpArg;
	tmpArg  = (struct threadArg_calc2DHist **)malloc( nThreads*sizeof(struct threadArg_calc2DHist*));

	for (int i=0; i < nThreads; i++) {
		tmpArg[i] = (struct threadArg_calc2DHist*) malloc(sizeof(struct threadArg_calc2DHist));
		tmpArg[i]->x    	= &edgePts_x[0];
		tmpArg[i]->y    	= &edgePts_y[0];
		tmpArg[i]->grad         = &edgePts_grad[0];
		tmpArg[i]->ori          = &edgePts_ori[0];
		tmpArg[i]->origin_x    	= origin_x;
		tmpArg[i]->origin_y    	= origin_y;		
		tmpArg[i]->width        = width_polar;
		tmpArg[i]->height       = height_polar;
		tmpArg[i]->rThetaPlot_ptr  = tmpRThetaPlot_ptr[i];
		tmpArg[i]->startInd     = i*threadWidth;
		tmpArg[i]->endInd        = (i+1)*threadWidth;
		pthread_create(&pth[i],NULL,threadFn_calc2DHist,(void*)tmpArg[i]);	
	}

	//Wait for the other threads to end
	for(int i=0; i< nThreads; i++){
		pthread_join(pth[i],NULL);
	}
	
	
	
	//for the remaining pixels
	for(int ind = nThreads*threadWidth; ind <edgePts_x.size(); ind++){
		calc2DHist(origin_x,
			   origin_y,
			   edgePts_x[ind],
			   edgePts_y[ind],
			  *(edgeGrad_ptr + width_cart*edgePts_y[ind] + edgePts_x[ind]),
			  *(edgeOri_ptr + width_cart*edgePts_y[ind] + edgePts_x[ind]),
			  width_polar,
			  height_polar,
			  rThetaPlot_ptr);
	}

	// Add all the polar edge maps & rescale!
	for(int i=0; i < height_polar; i++){
		for(int j=0; j < width_polar; j++){
			double sum=*(rThetaPlot_ptr + width_polar*i + j);
			for(int  threadId=0; threadId < nThreads; threadId++)
				sum += *(tmpRThetaPlot_ptr[threadId] + width_polar*i + j); 					
			*(rThetaPlot_ptr + width_polar*i + j) = sum; 
			maxIntensity = (maxIntensity < sum) ? sum: maxIntensity; 	
		}
	}

	// rescale the max value of the theta plot to lie betweeen 0 and 1
	for(int i=0; i < height_polar; i++)
		for(int j=0; j < width_polar; j++)
			*(rThetaPlot_ptr + width_polar*i + j) /= maxIntensity; 


	//release memory of the temporary pointers!
	for(int i=0; i<nThreads; i++){
		free(tmpRThetaPlot_ptr[i]);
		free(tmpArg[i]);		
	}
	
       fprintf(stdout,"\n"); fflush(stdout);
}


